import typing
import warnings
from inspect import getdoc
from itertools import count
from typing import List, Optional, Type, get_type_hints

from aiohttp.web import Response, json_response
from aiohttp.web_app import Application
from pydantic import RootModel

from ..injectors import _parse_func_signature
from ..utils import is_pydantic_base_model
from ..view import PydanticView, is_pydantic_view
from . import docstring_parser
from .definition import (
    AIOHTTP_HAS_APP_KEY,
    key_apps_to_expose,
    key_index_template,
    key_title_spec,
    key_version_spec,
    key_security,
)
from .struct import OpenApiSpec3, OperationObject, PathItem
from .typing import is_status_code_type

_APP_KEY_NOT_SET = object()


class _OASResponseBuilder:
    """
    Parse the type annotated as returned by a function and
    generate the OAS operation response.
    """

    def __init__(self, oas: OpenApiSpec3, oas_operation, status_code_descriptions):
        self._oas_operation = oas_operation
        self._oas = oas
        self._status_code_descriptions = status_code_descriptions

    def _handle_pydantic_base_model(self, obj):
        if is_pydantic_base_model(obj):
            response_schema = obj.model_json_schema(
                ref_template="#/components/schemas/{model}"
            ).copy()
            if def_sub_schemas := response_schema.pop("$defs", None):
                self._oas.components.schemas.update(def_sub_schemas)
            return response_schema
        return {}

    def _handle_list(self, obj):
        if typing.get_origin(obj) is list:
            return {
                "type": "array",
                "items": self._handle_pydantic_base_model(typing.get_args(obj)[0]),
            }
        return self._handle_pydantic_base_model(obj)

    def _handle_status_code_type(self, obj):
        if is_status_code_type(typing.get_origin(obj)):
            status_code = typing.get_origin(obj).__name__
            if status_code != "default":
                status_code = status_code[1:]
            self._oas_operation.responses[status_code].content = {
                "application/json": {
                    "schema": self._handle_list(typing.get_args(obj)[0])
                }
            }
            desc = self._status_code_descriptions.get(status_code)
            if desc:
                self._oas_operation.responses[status_code].description = desc

        elif is_status_code_type(obj):
            status_code = obj.__name__
            if status_code != "default":
                status_code = status_code[1:]
            self._oas_operation.responses[status_code].content = {}
            desc = self._status_code_descriptions.get(status_code)
            if desc:
                self._oas_operation.responses[status_code].description = desc

    def _handle_union(self, obj):
        if typing.get_origin(obj) is typing.Union:
            for arg in typing.get_args(obj):
                self._handle_status_code_type(arg)
        self._handle_status_code_type(obj)

    def build(self, obj):
        self._handle_union(obj)


def _add_http_method_to_oas(
    oas: OpenApiSpec3, oas_path: PathItem, http_method: str, view: Type[PydanticView]
):
    http_method = http_method.lower()
    oas_operation: OperationObject = getattr(oas_path, http_method)
    handler = getattr(view, http_method)
    path_args, body_args, qs_args, header_args, defaults = _parse_func_signature(
        handler, unpack_group=True
    )
    description = getdoc(handler)
    if description:
        oas_operation.description = docstring_parser.operation(description)
        oas_operation.tags = docstring_parser.tags(description)
        oas_operation.security = docstring_parser.security(description)
        oas_operation.deprecated = docstring_parser.deprecated(description)
        status_code_descriptions = docstring_parser.status_code(description)
    else:
        status_code_descriptions = {}

    if body_args:
        body_schema = (
            next(iter(body_args.values()))
            .model_json_schema(ref_template="#/components/schemas/{model}")
            .copy()
        )
        if def_sub_schemas := body_schema.pop("$defs", None):
            oas.components.schemas.update(def_sub_schemas)

        oas_operation.request_body.content = {
            "application/json": {"schema": body_schema}
        }

    indexes = count()
    for args_location, args in (
        ("path", path_args.items()),
        ("query", qs_args.items()),
        ("header", header_args.items()),
    ):
        for name, type_ in args:
            i = next(indexes)
            oas_operation.parameters[i].in_ = args_location
            oas_operation.parameters[i].name = name

            attrs = {"__annotations__": {"root": type_}}
            if name in defaults:
                attrs["root"] = defaults[name]
                oas_operation.parameters[i].required = False
            else:
                oas_operation.parameters[i].required = True

            attr_schema = type(name, (RootModel,), attrs).model_json_schema(
                ref_template="#/components/schemas/{model}"
            )
            if def_sub_schemas := attr_schema.pop("$defs", None):
                oas.components.schemas.update(def_sub_schemas)
            oas_operation.parameters[i].schema = attr_schema


    return_type = get_type_hints(handler).get("return")
    if return_type is not None:
        _OASResponseBuilder(oas, oas_operation, status_code_descriptions).build(
            return_type
        )


def generate_oas(
    apps: List[Application],
    version_spec: Optional[str] = None,
    title_spec: Optional[str] = None,
    security: Optional[dict] = None,
) -> dict:
    """
    Generate and return Open Api Specification from PydanticView in application.
    """
    oas = OpenApiSpec3()

    if version_spec is not None:
        oas.info.version = version_spec

    if title_spec is not None:
        oas.info.title = title_spec

    for app in apps:
        for resources in app.router.resources():
            for resource_route in resources:
                if not is_pydantic_view(resource_route.handler):
                    continue

                view: Type[PydanticView] = resource_route.handler
                info = resource_route.get_info()
                path = oas.paths[info.get("path", info.get("formatter"))]
                if resource_route.method == "*":
                    for method_name in view.allowed_methods:
                        _add_http_method_to_oas(oas, path, method_name, view)
                else:
                    _add_http_method_to_oas(oas, path, resource_route.method, view)

    if security:
        oas.components.security_schemes.update(security)

    return oas.spec


def _app_key_or_string(app, app_key, str_key):
    if app_key in app:
        return app[app_key]
    if AIOHTTP_HAS_APP_KEY:
        key_name = "key_" + str_key.replace(" ", "_")
        warnings.warn(f"Use from aiohttp_pydantic.oas.definition import {key_name}; app[{key_name}] = ... "
                      f"instead of app[{str_key}] = ...", DeprecationWarning, stacklevel=4)
    return app[str_key]


async def get_oas(request):
    """
    View to generate the Open Api Specification from PydanticView in application.
    """
    apps = _app_key_or_string(request.app, key_apps_to_expose, "apps to expose")
    version_spec = _app_key_or_string(request.app, key_version_spec, "version spec")
    title_spec = _app_key_or_string(request.app, key_title_spec, "title spec")
    security = _app_key_or_string(request.app, key_security, "security")
    return json_response(generate_oas(apps, version_spec, title_spec, security))


async def oas_ui(request):
    """
    View to serve the swagger-ui to read open api specification of application.
    """
    template = _app_key_or_string(request.app, key_index_template, "index template")

    return Response(
        text=template.render(
            {
                "openapi_spec_url": request.app.router['spec'].canonical,
                "static_url": request.app.router['static'].canonical,
            }
        ),
        content_type="text/html",
        charset="utf-8",
    )
